#if defined(ARDUINO) && ARDUINO >= 100
  #include "Arduino.h"
#elif defined(ENERGIA)
  #include "Energia.h"
#else
  #include "WProgram.h"
#endif

#include <avr/interrupt.h>
#include <avr/power.h>
#include <avr/sleep.h>

#include "Sleep.h"

static inline void power_down() {
  // power off ADC, Timer 0 and 1, serial interface
  ACSR |= 1 << ACD;
  ADCSRA &= ~(1 << ADEN);
  
  power_all_disable();
}

static inline void power_up() {
  power_all_enable();

  ACSR &= ~(1 << ACD);
  ADCSRA |= 1 << ADEN;
}

void sleepForever() {
  set_sleep_mode(SLEEP_MODE_PWR_DOWN);

  power_down();

  // On Attiny brown-out detection is disabled by the fuses
  sleep_mode();

  power_up();
}

#define wdt_reset() __asm__ __volatile__ ("wdr")

static void wdt_disable() {
  // Disable Watchdog
  noInterrupts();
  
  wdt_reset();
  /* Clear WDRF in MCUSR */
  MCUSR = 0;
  // Do not optimize!
  WDTCSR |= (1 << WDCE) | (1 << WDE);
  WDTCSR = 0;

  interrupts();
}

ISR(WDT_vect) { }

static void wdt_enable_8s() {
  // Normal wdt_enable() does not work properly
  noInterrupts();
  
  // Clear the reset flag
  MCUSR &= ~(1 << WDRF);
  
  /*
   * In order to change WDE or the prescaler, we need to
   * set WDCE (This will allow updates for 4 clock cycles).
   */
  WDTCSR |= (1 << WDCE) | (1 << WDE);

  // set new watchdog timeout prescaler value
  // 8.0 seconds
  WDTCSR = 1 << WDP0 | 1 << WDP3;
  
  // Enable the WD interrupt (note no reset)
  WDTCSR |= _BV(WDIE);

  interrupts();
}

static void wdt_enable_500ms() {
  // Normal wdt_enable() does not work properly
  noInterrupts();
  
  // Clear the reset flag
  MCUSR &= ~(1 << WDRF);
  
  /*
   * In order to change WDE or the prescaler, we need to
   * set WDCE (This will allow updates for 4 clock cycles).
   */
  WDTCSR |= (1 << WDCE) | (1 << WDE);

  // set new watchdog timeout prescaler value
  // 500ms
  WDTCSR = 1 << WDP0 | 1 << WDP2;
  
  // Enable the WD interrupt (note no reset)
  WDTCSR |= _BV(WDIE);

  interrupts();
}

static void wdt_enable_250ms() {
  // Normal wdt_enable() does not work properly
  noInterrupts();
  
  // Clear the reset flag
  MCUSR &= ~(1 << WDRF);
  
  /*
   * In order to change WDE or the prescaler, we need to
   * set WDCE (This will allow updates for 4 clock cycles).
   */
  WDTCSR |= (1 << WDCE) | (1 << WDE);

  // set new watchdog timeout prescaler value
  // 500ms
  WDTCSR = 1 << WDP0 | 1 << WDP2;
  
  // Enable the WD interrupt (note no reset)
  WDTCSR |= _BV(WDIE);

  interrupts();
}

void sleep500ms() {
  wdt_enable_500ms();

  sleepForever();

  wdt_disable();
}

void sleep250ms() {
  wdt_enable_250ms();

  sleepForever();

  wdt_disable();
}

void sleep(unsigned int secs) {
  set_sleep_mode(SLEEP_MODE_PWR_DOWN);

  // On Attiny brown-out detection is disabled by the fuses
  power_down();

  while(secs >= 8) {
    wdt_enable_8s();
    
    sleep_mode();
    secs -= 8;
  }

  wdt_disable();
  
  power_up();
}
